#
# Copyright (c) 2021-2022, 2025 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Wed Dec 15 2021
#
"""Client side component interface."""
import types

from . import event
from . import pmt


# --- load -----------------------------------------------------------------
#
def load(server, client, *args):
  """Create a component client instance."""

  # parse args
  args = list(args)
  cargs = []
  while args:
    arg = args.pop(0)
    if arg == '-h':
      # usage
      def load(client=client, *args): pass
      load.__doc__ = (
        'Create a component client instance.\n\n'
        'Available %s options:\n' % client
      )
      for i in server.options(client) + [{
        'option': '-h', 'descr': 'show this help'
      }]:
        load.__doc__ += '  %-26s %s\n' % (i['option'], i['descr'])

      help(load)
      return
    else:
      # remote arguments
      cargs.append(str(arg))

  return Handle(server, client, cargs)


# === Handle ===============================================================
#
class Handle:
  """Generic component client."""

  __handles = {}


  # --- __new__ ------------------------------------------------------------
  #
  def __new__(cls, server, client, cargs):
    """Load a component client and return a populated Handle instance."""

    # remote load component and return any already loaded handle type
    token = server.dlopen(client, cargs)
    H = cls.__handles.get((server, token), None)
    if H: return super().__new__(H)

    # create a new Handle subclass. This is mostly to accomodate with
    # the interactive help system, as the latter does not document
    # methods added on instances. So the subclass itself is populated
    # with services and ports.
    H = type(client, (Handle,), {})
    H.__doc__ = client + ' component client.'
    H.__server = server
    H.__token = token

    # get info
    info = server.info(token)

    # create services
    for s, desc in info['services'].items():
      setattr(H, pmt.sanitize(s), H.__partial_request(s, desc))

    # create ports
    for p, desc in info['ports'].items():
      setattr(H, pmt.sanitize(p), H.__partial_port(p, desc))

    # create publishers
    for p, desc in info.get('pubs', {}).items():
      setattr(H, pmt.sanitize(p), H.__partial_pub(p, desc))

    # record type
    cls.__handles[(server, token)] = H

    # return instance
    return super().__new__(H)


  # --- __repr__ -----------------------------------------------------------
  #
  @classmethod
  def __repr__(cls):
    return '<%s component%s%s on %s>' % (
      cls.__name__,
      (' (%s instance)' % cls.__token) \
      if cls.__token and cls.__token != cls.__name__ else '',
      ' [unloaded]' if cls.__token is None else '',
      cls.__server.host
    )


  # --- close --------------------------------------------------------------
  #
  @classmethod
  def close(cls, server, token):
    """Forget about this component."""

    for t in [t for s, t in cls.__handles
              if s == server and (t == token or token is ...)]:
      H = cls.__handles.pop((server, t), None)
      if H: H.__token = None


  # --- _request -----------------------------------------------------------
  #
  @classmethod
  def __request(cls, service, desc, defarg, *args, **kwargs):
    """Send a generic request to a component."""

    if cls.__token is None: raise RuntimeError('unloaded component')

    # defaults
    params = {}
    cb = None
    wait = event.Status.done
    timeout = kwargs.pop('timeout', None)

    # return metadata or help
    if kwargs.pop('args', False): return desc
    if kwargs.pop('help', False): return help(getattr(cls, service))

    # get callback, default to async if set
    try: cb = kwargs.pop('callback')
    except KeyError: pass
    else:
      if not callable(cb):
        raise TypeError('callback must be callable')
      wait = event.Status.sent

    # get async options
    if kwargs.pop('ack', False): wait = event.Status.sent
    if kwargs.pop('send', False): wait = event.Status.none
    if kwargs.pop('oneway', False):
      wait = event.Status.none
      params['oneway'] = 1

    # create json argument
    if not args and not kwargs:
      args = []
      kwargs = pmt.iscanargs(desc['input'], desc['meta'])
    elif len(args) == 1 and isinstance(args[0], dict):
      kwargs.update(args[0])
      args = []
    else:
      args = list(args)

    params['input'] = pmt.arg2json(args, kwargs, defarg, desc['input'])

    # send remotely (note: may also process unrelated new events)
    r = cls.__server.send(cls.__token, service, params)
    if r is None: return None

    # finalize
    r.callback = cb
    r.wait(status = wait, timeout = timeout)

    return r.result if wait == event.Status.done else r


  # --- _port --------------------------------------------------------------
  #
  @classmethod
  def __port(cls, port, desc, *args):
    """Read a generic port from a component."""

    if cls.__token is None: raise RuntimeError('unloaded component')

    if desc['simple']:
      if len(args) != 0:
        raise TypeError('%s.%s() expects no argument' % (cls.__token, port))
    else:
      if len(args) != 1:
        raise TypeError(
          '%s.%s() expects a port handle name argument' % (cls.__token, port))
      port += '/' + str(args[0])

    return cls.__server.read(cls.__token, port)


  # --- __pub --------------------------------------------------------------
  #
  @classmethod
  def __pub(cls, port, desc, path):
    """Open or publish a new port from a component."""

    if cls.__token is None: raise RuntimeError('unloaded component')

    # keep token as a reference to the component one - this is only useful
    # if the component is unloaded and token is reset to None
    token = lambda: cls.__token

    client = cls.__token
    server = cls.__server
    fqpn = server.pub(cls.__token, port, 'open', str(path))

    # define publisher object
    class Publisher:
      """Create port publisher."""
      __doc__ = '%s (%s.%s) port publisher.' % (fqpn, client, port)

      @property
      def name(self): return fqpn

      def __repr__(self):
        return '<%s port (%s.%s instance)%s on %s>' % (
          fqpn, client, port, ' [unloaded]' if token() is None else '',
          server.host
        )

      def data(self):
        """Get last published data."""
        if token() is None: raise RuntimeError('unloaded component')
        return server.pub(token(), port, 'data', fqpn)

      def close(self):
        """Close port instance."""
        if token() is None: raise RuntimeError('unloaded component')
        server.pub(token(), port, 'close', fqpn)

      def __call__(self, *args, **kwargs):
        if token() is None: raise RuntimeError('unloaded component')

        # create json argument
        if not args and not kwargs:
          args = []
          kwargs = pmt.iscanargs(desc['data'], desc['meta'])
        elif len(args) == 1 and isinstance(args[0], dict):
          kwargs.update(args[0])
          args = []
        else:
          args = list(args)

        server.pub(token(), port, 'write', fqpn, {
          'data': pmt.arg2json(args, kwargs, {}, desc['data'])
        })

      __call__.__doc__ = (
        'Publish "%s" port.\n\n' % fqpn +
        'Input dictionary:' + pmt.meta2doc(desc['data'], desc['meta'])
      )

    Publisher.__name__ = port
    return Publisher()


  # --- __partial_request --------------------------------------------------
  #
  @classmethod
  def __partial_request(cls, service, desc):
    """Return a closure for invoking a specialized request sending
    function.
    """

    defarg = pmt.meta2defarg(desc['meta'])

    def partial(cls, *args, **kwargs):
      return cls.__request(service, desc, defarg, *args, **kwargs)

    partial.__name__ = service
    partial.__doc__ = (
      'Send ' + service + ' request.\n\n'
      'This returns a dictionary containing the output result or a\n'
      'request handle, depending on the options.\n\n'
      'Keywords arguments:\n' +
      '\n'.join(['  %-16s %s' % (k,v) for k,v in (
        ('send=True', 'Return a request handle immediately after'),
        ('','sending the request.'),

        ('ack=True', 'Return a request handle after sending the'),
        ('','request and it has been acknowledged remotely.'),

        ('oneway=True', 'Return None immediately after sending the request.'),

        ('callback=...', 'Arrange for the given callable to be called after'),
        ('', 'each status update (sent, done or error) of the'),
        ('', 'request. A request handle is returned.'),

        ('timeout=...', 'Maximum amount of time to wait.'),

        ('args=True', 'Return a dictionary with service metada.'),
        ('help=True', 'Show this help.'),
      )]) +
      '\n\nInput dictionary:' +
      pmt.meta2doc(desc['input'], desc['meta']) +
      '\n\nOutput dictionary:' +
      pmt.meta2doc(desc['output'], desc['meta'])
    )

    return partial


  # --- __partial_port -----------------------------------------------------
  #
  @classmethod
  def __partial_port(cls, port, desc):
    """Return a closure for invoking a specialized port reading
    function.
    """

    def partial(cls, *args, **kwargs):
      return cls.__port(port, desc, *args, **kwargs)

    partial.__name__ = port
    partial.__doc__ = (
      'Read "%s" port.\n\n' % port +
      'Output dictionary:' +
      pmt.meta2doc(desc['data'], {})
    )

    return partial


  # --- __partial_pub ------------------------------------------------------
  #
  @classmethod
  def __partial_pub(cls, port, desc):
    """Return a closure for creating a publisher handle.
    """

    def partial(cls, path):
      return cls.__pub(port, desc, path)

    partial.__name__ = port
    partial.__doc__ = 'Publish "%s" port.' % port

    return partial
